import os
import f90nml
import numpy as np


class Fun3dStaticAeroelasticDriver:
    def __init__(self, number_of_modes, extra_command_line_args=''):
        self.num_modes = number_of_modes

        self.modal_disp = np.zeros(self.num_modes)

        self.mpirun = 'mpiexec_mpt'
        self.nodet_mpi = 'nodet_mpi'
        self.extra_command_line_args = extra_command_line_args

        # Aitken variables
        self.theta = 0.5
        self.theta_min = 0.01
        self.theta_max = 1.0
        self.previous_update = np.zeros(self.num_modes)

        self._compute_modal_stiffnesses()

    def _compute_modal_stiffnesses(self):
        self.modal_stiffness = np.zeros(self.num_modes)

        nml = f90nml.read('moving_body.input')
        for mode in range(self.num_modes):
            mass = nml['aeroelastic_modal_data']['gmass'][0][mode]
            freq = nml['aeroelastic_modal_data']['freq'][0][mode]
            self.modal_stiffness[mode] = mass * freq**2.0

    def run(self, number_of_coupling_steps):
        for step in range(number_of_coupling_steps):
            print(f'Coupling step {step}: theta = {self.theta}')
            self.set_restart_read(step)
            self.set_modal_displacements()
            self.run_fun3d(step)
            self.compute_new_modal_displacements(step)

    def set_restart_read(self,step):
        restart_read = 'on_nohistorykept' if step == 0 else 'on'
        nml = f90nml.read('fun3d.nml')
        nml['code_run_control']['restart_read'] = restart_read
        nml.write('fun3d.nml', force=True)

    def set_modal_displacements(self):
        nml = f90nml.read('moving_body.input')
        for mode in range(self.num_modes):
            nml['aeroelastic_modal_data']['gdisp0'][0][mode] = self.modal_disp[mode]
        nml.write('moving_body.input', force=True)

    def run_fun3d(self, step):
        command_line_args = f'--aeroelastic_internal {self.extra_command_line_args}'
        command = f'{self.mpirun} {self.nodet_mpi} {command_line_args} > static_{step}.out'
        os.system(command)

    def compute_new_modal_displacements(self, step):
        new_disps = np.zeros(self.num_modes)

        for mode in range(self.num_modes):
            filename = f'aehist_body1_mode{mode+1}.dat'
            modal_data = np.loadtxt(filename, skiprows=3)
            generalized_force = modal_data[-1,3]
            new_disps[mode] = generalized_force / self.modal_stiffness[mode]
        self._aitken_relax(step, new_disps)

    def _aitken_relax(self, step, new_disps):
        update = new_disps - self.modal_disp
        if step > 0:
            norm2 = (np.linalg.norm(update - self.previous_update)**2.0)

            # Only update theta if the displacements changed
            if norm2 > 1e-13:
                self.theta *= 1.0 - (update - self.previous_update).dot(update)/norm2
                self.theta = np.max((np.min((self.theta,self.theta_max)),self.theta_min))

        self.previous_update = update.copy()
        self.modal_disp += self.theta * update


if __name__ == '__main__':
    number_of_coupling_steps = 20
    nmodes = 2
    extra_args = '--gamma 1.136'
    driver = Fun3dStaticAeroelasticDriver(nmodes, extra_args)

    driver.run(number_of_coupling_steps)
